#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#define MODULE
#define __KERNEL__
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include "linux/netdevice.h"
#include <linux/fs.h>
#include <linux/ip.h>
#include <asm/signal.h>
#include <asm/unistd.h>
#include <asm/semaphore.h>


/* A wrapper which does next to nothing at
 * at present, but may help for compatibility
 * with future versions of Linux */
#include <linux/wrapper.h>
			     
/* Our own ioctl numbers */
#include "include/common.h"
#include "include/module.h"
//#include "include/flame_hook.h"
//#include "frey_hook.h" /* XXX anagnost */
#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
#include <asm/uaccess.h>  /* for get_user and put_user */
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
typedef struct net_device ndevice;
#include "linux/devfs_fs_kernel.h"
#else
typedef struct device ndevice;
#endif

/* Device Declarations ******************************** */


/* The name for our device, as it will appear in 
 * /proc/devices */
#define DEVICE_NAME "flamedev"

/* Is the device open right now? Used to prevent 
 * concurent access into the same device */
static int Device_Open = 0;

/* our externed variables in net/core/dev.c */
extern int	mypmodules_add();
extern int	frey_hookmod();
//extern int	frey_hookmod();
extern unsigned long	module_exception();

/* our polling device */
static ndevice *sampling_dev = NULL;
char *devname = "eth4";
// need this for linux 2.4.x
MODULE_LICENSE("GPL");

struct packet_type *orig_ip_packet = NULL;

int
flame_ip_rcv(struct sk_buff *skb, struct net_device *dev, 
            struct packet_type *pt)
{
 struct iphdr *iph = skb->nh.iph;
 frey_hookmod(skb);
 if(!strcmp(skb->dev->name,devname))
     printk("lala %u.%u.%u.%u -> %u.%u.%u.%u\n", NIPQUAD(iph->saddr),
            NIPQUAD(iph->daddr));
 
/*
 if( ntohs(iph->id) == 0 && ntohl(iph->saddr) == 0){
     ;
 }
*/
 return(orig_ip_packet->func(skb, dev, pt));
}


struct packet_type flame_ip_packet =
{
    __constant_htons(ETH_P_IP),
    NULL,
    flame_ip_rcv,
    NULL,
    NULL
};

// now for the exception handler stuff
struct idt_entry {
      unsigned short offset_low, selector;
      unsigned char flag, reserved;
      unsigned short offset_high;
};

struct desc_struct {
        unsigned long a,b;
};

unsigned long (*old_do_page_fault)();
unsigned long old_func;
unsigned long old_addr;

// call module_exception and return if its ours
// We DO NOT want to call module_exception on a OK page fault, only on a BAD
// one
// modules will cause page faults to load it into memory
// otherwise, just call the real do_page_fault()
asmlinkage void my_do_page_fault(struct pt_regs *regs, unsigned long error_code)
{
    struct task_struct *tsk;
    struct mm_struct *mm;
    struct vm_area_struct *vma;
    unsigned long address;
    unsigned long newip;

    __asm__("movl %%cr2,%0":"=r" (address));

    // It's safe to allow irq's after cr2 has been saved 
    if (regs->eflags & X86_EFLAGS_IF)
        local_irq_enable();
    
    tsk = current;
    if (address >= TASK_SIZE && !(error_code & 5))
        goto page_fault;

    mm = tsk->mm;

    if(in_interrupt() || !mm)
        goto check_fault;

    down_read(&mm->mmap_sem);
    vma = find_vma(mm, address);
    up_read(&mm->mmap_sem);

    if(!vma || vma->vm_start <= address || vma->vm_flags & VM_GROWSDOWN)
        goto page_fault;
    
  check_fault:
    if(module_exception){
        newip = module_exception(regs->eip);
        if(newip) {
            regs->eip=newip;
            return;
        }
    }
    
  page_fault:
    
    old_do_page_fault(regs,error_code);
}


/* This function is called whenever a process attempts 
 * to open the device file */
static int device_open(struct inode *inode, 
                       struct file *file)
{
#ifdef DEBUG
    printk ("device_open(%p)\n", file);
#endif
  /* We don't want to talk to two processes at the 
   * same time */
  if (Device_Open)
    return -EBUSY;

  /* If this was a process, we would have had to be 
   * more careful here, because one process might have 
   * checked Device_Open right before the other one 
   * tried to increment it. However, we're in the 
   * kernel, so we're protected against context switches.
   *
   * This is NOT the right attitude to take, because we
   * might be running on an SMP box, but we'll deal with
   * SMP in a later chapter.
   */ 

  Device_Open++;

  MOD_INC_USE_COUNT;

  return 0;
}


/* This function is called when a process closes the 
 * device file. It doesn't have a return value because 
 * it cannot fail. Regardless of what else happens, you 
 * should always be able to close a device (in 2.0, a 2.2
 * device file could be impossible to close). */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static int device_release(struct inode *inode, 
                          struct file *file)
#else
static void device_release(struct inode *inode, 
                           struct file *file)
#endif
{
#ifdef DEBUG
    printk ("device_release(%p,%p)\n", inode, file);
#endif 
  /* We're now ready for our next caller */
  Device_Open --;

  MOD_DEC_USE_COUNT;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  return 0;
#endif
}



/* This function is called whenever a process which 
 * has already opened the device file attempts to 
 * read from it. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t device_read(
    struct file *file,
    char *buffer, /* The buffer to fill with the data */   
    size_t length,     /* The length of the buffer */
    loff_t *offset) /* offset to the file */
#else
static int device_read(
    struct inode *inode,
    struct file *file,
    char *buffer,   /* The buffer to fill with the data */ 
    int length)     /* The length of the buffer 
                     * (mustn't write beyond that!) */
#endif
{
  /* Number of bytes actually written to the buffer */
  int bytes_read = 0;

#ifdef DEBUG
  printk("device_read(%p,%p,%d)\n",
    file, buffer, length);
#endif

#ifdef DEBUG
   printk ("Read %d bytes, %d left\n",
     bytes_read, length);
#endif

  return bytes_read;
}


/* This function is called when somebody tries to 
 * write into our device file. */ 
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t device_write(struct file *file,
                            const char *buffer,
                            size_t length,
                            loff_t *offset)
#else
static int device_write(struct inode *inode,
                        struct file *file,
                        const char *buffer,
                        int length)
#endif
{

#ifdef DEBUG
  printk ("device_write(%p,%s,%d)",
    file, buffer, length);
#endif

  /* Again, return the number of input characters used */
  return 0;
}


/* This function is called whenever a process tries to 
 * do an ioctl on our device file. We get two extra 
 * parameters (additional to the inode and file 
 * structures, which all device functions get): the number
 * of the ioctl called and the parameter given to the 
 * ioctl function.
 *
 * If the ioctl is write or read/write (meaning output 
 * is returned to the calling process), the ioctl call 
 * returns the output of this function.
 */
int device_ioctl(
    struct inode *inode,
    struct file *file,
    unsigned int ioctl_num,/* The number of the ioctl */
    unsigned long ioctl_param) /* The parameter to it */
{
  int i;

  struct flamedev_io *d;
    int error = 0;
    char mod_request[MAXMSGLEN];
    char * result;
    int mod_id = -1;

    /* Switch according to the ioctl called */
  switch (ioctl_num) {

  case MODQUERY:

      d = (struct flamedev_io *) ioctl_param;
      memset(mod_request, '\0', MAXMSGLEN);
      error = copy_from_user(&mod_request, d->request, MAXMSGLEN - 1);
#ifdef DEBUG
      printk("requested module = %s\n",mod_request);
#endif      
      i=0;
      
//      while(pmodules_name[i] != NULL && i < MAX_PMODULES) {
      while(i < MAX_PMODULES) {
          if(pmodules_name[i] != NULL && !strcmp(pmodules_name[i],mod_request)) {
              mod_id = i;
              break;
          }
          i++;
      }
      
      // check if module exists
      if(mod_id < 0) {
          printk("Cannot find module=[%s]!!\n", mod_request);
          return -1;
      }
            
      
      // query our module!
      if(pmodules[mod_id] != NULL && pmodules_query[mod_id] != NULL) {
#ifdef DEBUG
          printk("Do query at %x\n",(unsigned int)pmodules_query[mod_id]);
#endif
          result = pmodules_query[mod_id](mod_request);
          result = pmodules_query_return[mod_id]();
          error = copy_to_user(d->response, result, MAXMSGLEN - 1);
      } else {
          printk("ERROR finding module for mod_id=%d\n",mod_id);
          // do something useful here
          error = -1;
      }
      break;
      
  }  
      return 0;
}



/* Module Declarations *************************** */
/* This structure will hold the functions to be called 
 * when a process does something to the device we 
 * created. Since a pointer to this structure is kept in 
 * the devices table, it can't be local to
 * init_module. NULL is for unimplemented functions. */
// new C99 syntax
struct file_operations Fops = {
    .read  =  (void *)device_read, 
    .write = (void *)device_write,
    .ioctl =  device_ioctl,   /* ioctl */
    .open  =  device_open,
    .release =  device_release  /* a.k.a. close */
};


/* Initialize the module - Register the character device */
int init_module()
{
    int ret_val=0;
    int i=0;
    ndevice *dev = (ndevice *) NULL;
    struct desc_struct *base;
    unsigned char idtr[6];  // IDTR is 48 bits
    unsigned char *p;
    unsigned long target;
    
// see intel manual vol 3

    // perhaps this should be a command line option
    dev = dev_get_by_name((const char *)devname);
    
    if(dev==NULL) {      
        printk("Where is eth1?\n Cannot setup polling!!  \nPlease unload
 this module!!\n");
        return -1;
    }
    
    // save our ptr for unloading
    sampling_dev = dev;

    dev_add_pack(&flame_ip_packet);
    orig_ip_packet = flame_ip_packet.next;
    dev_remove_pack(orig_ip_packet); 

    // exception handler stuff
    //get idt address
    __asm__ volatile("sidt %0": "=m"(idtr));
    base = (struct desc_struct *)(*(unsigned long *)(&idtr[2]));
    base += 14;
    target = (base->a & 0xffff) | (base->b &0xffff0000);
    p = (unsigned char *) target;
    if(*p == 0x68) { // instr code of pushl
        printk("address is %x\n",*(unsigned int*)(p+1));
        old_addr = (unsigned long)(p+1);
        
        (unsigned long) old_do_page_fault = *(unsigned long*)(p+1);
        *(unsigned long*)(p+1) = (unsigned long) my_do_page_fault;
        printk("address of old_do_page_fault = %x\n",(unsigned int) old_do_page_fault);
    }
    
#define module_register_chrdev devfs_register_chrdev
#define module_unregister_chrdev devfs_unregister_chrdev

/* Register the character device (atleast try) */
    ret_val = module_register_chrdev(MAJOR_NUM, 
                                     DEVICE_NAME,
                                     &Fops);
    
/* Negative values signify an error */
    if (ret_val < 0) {
        printk ("%s failed with %d\n",
                "Sorry, registering the character device ",
                ret_val);
        return ret_val;
    }
    printk("==== FLAMEDEV =======\n");
//    checkfault = module_exception;
    
// default cycles allowed
    for(i=0;i<MAX_PMODULES;i++) {
        cycles_allowed[i] = 50000;
        pmodules[i] = NULL;
        pmodules_name[i] = NULL;
    }
    return 0;
}


/* Cleanup - unregister the appropriate file from /proc */
void cleanup_module()
{
int ret;
int i;
printk ("Unloading FLAMEDEV...\n");

//checkfault = NULL;

dev_add_pack(orig_ip_packet);
dev_remove_pack(&flame_ip_packet); 
*(unsigned long *) old_addr = (unsigned long) old_do_page_fault;

// empty module list
for(i=0;i<MAX_PMODULES;i++) {
pmodules[i]=NULL;
pmodules_name[i]=NULL;
}

/* Unregister the device */
ret = module_unregister_chrdev(MAJOR_NUM, DEVICE_NAME);

/* If there's an error, report it */ 
if (ret < 0)
printk("Error in module_unregister_chrdev: %d\n", ret);
}  


/* end of file */
